"""Helper APIs for interaction with Google Fonts glyphsets.

Provides APIs to interact with font subsets, codepoints for font or subset.
"""
import codecs
import contextlib
import re
import sys
import warnings
# importlib.resources.files only available since Python>3.9,
# temporarily use backport
from importlib_resources import files
from fontTools import ttLib

from . import subsets

nam_files = files("gfsubsets.data")


class NamFileDict(dict):
    cache = {}

    def __getitem__(self, subset):
        if subset not in self.cache:
            self.cache[subset] = self.read_namfile(subset)
        return self.cache[subset]

    def read_namfile(self, subset):
        cps = set()
        ref = nam_files.joinpath(subset + "_unique-glyphs.nam")
        if not ref.is_file():
            warnings.warn(f"No such subset '{subset}'")
            return cps
        with ref.open("r", encoding="utf-8") as f:
            for line in f:
                if line.startswith("#"):
                    continue
                match = _NAMELIST_CODEPOINT_REGEX.match(line[2:7])
                if match is not None:
                    cps.add(int(match.groups()[0], 16))
        return cps


subset_codepoints = NamFileDict()

# Matches 4 or 5 hexadecimal digits that are uppercase at the beginning of the
# test string. The match is stored in group 0, e.g:
# >>> _NAMELIST_CODEPOINT_REGEX.match('1234X').groups()[0]
# '1234'
# >>> _NAMELIST_CODEPOINT_REGEX.match('1234A').groups()[0]
# '1234A'
_NAMELIST_CODEPOINT_REGEX = re.compile("^([A-F0-9]{4,5})")

_PLATFORM_ID_MICROSOFT = 3
_PLATFORM_ENC_UNICODE_BMP = 1
_PLATFORM_ENC_UNICODE_UCS4 = 10
_PLATFORM_ENCS_UNICODE = (_PLATFORM_ENC_UNICODE_BMP, _PLATFORM_ENC_UNICODE_UCS4)


def UnicodeCmapTables(font):
    """Find unicode cmap tables in font.

    Args:
      font: A TTFont.
    Yields:
      cmap tables that contain unicode mappings
    """
    for table in font["cmap"].tables:
        if (
            table.platformID == _PLATFORM_ID_MICROSOFT
            and table.platEncID in _PLATFORM_ENCS_UNICODE
        ):
            yield table


def CodepointsInFont(font_filename):
    """Returns the set of codepoints present in the font file specified.

    Args:
      font_filename: The name of a font file.
    Returns:
      A set of integers, each representing a codepoint present in font.
    """

    font_cps = set()
    with contextlib.closing(ttLib.TTFont(font_filename)) as font:
        for t in UnicodeCmapTables(font):
            font_cps.update(t.cmap.keys())

    return font_cps


def ListSubsets():
    """Returns a list of all subset names, in lowercase."""
    return subsets.SUBSETS


def SubsetsForCodepoint(cp):
    """Returns all the subsets that contains cp or [].

    Args:
      cp: int codepoint.
    Returns:
      List of lowercase names of subsets or [] if none match.
    """
    _subsets = []
    for subset in ListSubsets():
        cps = CodepointsInSubset(subset, unique_glyphs=True)
        if not cps:
            continue
        if cp in cps:
            _subsets.append(subset)
    return _subsets


def SubsetForCodepoint(cp):
    """Returns the highest priority subset that contains cp or None.

    Args:
      cp: int codepoint.
    Returns:
      The lowercase name of the subset, e.g. latin, or None.
    """
    _subsets = SubsetsForCodepoint(cp)
    if not _subsets:
        return None

    result = _subsets[0]
    for subset in sorted(_subsets):
        # prefer x to x-ext
        if result + "-ext" == subset:
            pass
        elif result == subset + "-ext":
            # prefer no -ext to -ext
            result = subset
        elif subset.startswith("latin"):
            # prefer latin to anything non-latin
            result = subset

    return result


def set_encoding_path(enc_path):
    raise NotImplementedError("set_encoding_path is no longer supported")


def CodepointsInSubset(subset, unique_glyphs=False):
    """Returns the set of codepoints contained in a given subset.

    Args:
      subset: The lowercase name of a subset, e.g. latin.
      unique_glyphs: Optional, whether to only include glyphs unique to subset.
    Returns:
      A set containing the glyphs in the subset.
    """
    cps = subset_codepoints[subset]

    if unique_glyphs:
        return cps

    # y-ext includes y
    # Except latin-ext which already has latin.
    if subset != "latin-ext" and subset.endswith("-ext"):
        cps |= subset_codepoints[subset[:-4]]

    # almost all subsets include latin.
    if subset not in ("khmer", "latin"):
        cps |= subset_codepoints["latin"]
    return cps


def SubsetsInFont(file_path, min_pct, ext_min_pct=None):
    """Finds all subsets for which we support > min_pct of codepoints.

    Args:
      file_path: A file_path to a font file.
      min_pct: Min percent coverage to report a subset. 0 means at least 1 glyph.
      25 means 25%.
      ext_min_pct: The minimum percent coverage to report a -ext
      subset supported. Used to admit extended subsets with a lower percent. Same
      interpretation as min_pct. If None same as min_pct.
    Returns:
      A list of 3-tuples of (subset name, #supported, #in subset).
    """
    all_cps = CodepointsInFont(file_path)
    # Remove control chars and whitespace chars from the font since they may
    # trigger greek-ext and latin-ext subsets to be enabled.
    # https://github.com/googlefonts/nam-files/issues/14#issuecomment-2076693612
    for codepoint in set([0x0000, 0x000D, 0x0020, 0x00A0]):
        if codepoint in all_cps:
            all_cps.remove(codepoint)

    results = []
    for subset in ListSubsets():
        subset_cps = CodepointsInSubset(subset, unique_glyphs=True)
        if not subset_cps:
            continue

        # Khmer includes latin but we only want to report support for non-Latin.
        if subset == "khmer":
            subset_cps -= CodepointsInSubset("latin")

        overlap = all_cps & subset_cps

        target_pct = min_pct
        if ext_min_pct is not None and subset.endswith("-ext"):
            target_pct = ext_min_pct

        if 100.0 * len(overlap) / len(subset_cps) > target_pct:
            results.append((subset, len(overlap), len(subset_cps)))

    return results



#!/usr/bin/env python3
#
# Copyright 2017-2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
# importlib.resources.files only available since Python>3.9,
# temporarily use backport
from importlib_resources import files

subset_files = [resource.name for resource 
                in files("gfsubsets.data").iterdir()
                if resource.name.endswith("_unique-glyphs.nam")]

SUBSETS = sorted(
    filename.replace("_unique-glyphs.nam", "")
    for filename in subset_files
)



from collections import defaultdict
from gfsubsets import ListSubsets, CodepointsInSubset

ALLOWED = set(
    [
        0x0,
        0x20,
        0x000D,  # CARRIAGE RETURN
        0x002D,  # HYPHEN-MINUS
        0x00A0,  # NO-BREAK SPACE
        0x25CC,  # DOTTED CIRCLE
        0x200B,  # ZERO WIDTH SPACE
        0x200C,  # ZERO WIDTH NON-JOINER
        0x200D,  # ZERO WIDTH JOINER
        0x2010,  # HYPHEN
    ]
)

# Latin kernel gets *everywhere*
ALLOWED |= set(range(0x30, 0x7F))

SKIP_SUBSETS = [
    "chinese-simplified",
    "chinese-traditional",
    "chinese-hongkong",
    "japanese",
    "yi",
    "meroitic",
]

# Grandfather existing problems; we just don't want to make things worse.
EXISTING = [
    0x001D,
    range(0x0021, 0x002D),
    range(0x002E, 0x0030),
    0x00A7,
    range(0x00AB, 0x00AE),
    range(0x00B0, 0x00B2),
    0x00B7,
    0x00BB,
    0x00D7,
    0x00F7,
    range(0x0102, 0x0104),
    range(0x0110, 0x0112),
    range(0x0128, 0x012A),
    0x0131,
    range(0x0168, 0x016A),
    range(0x01A0, 0x01A2),
    range(0x01AF, 0x01B1),
    0x02BC,
    0x02C7,
    0x02DA,
    range(0x0300, 0x030A),
    range(0x030B, 0x030D),
    range(0x0323, 0x0325),
    0x0329,
    range(0x0330, 0x0332),
    range(0x0374, 0x0376),
    range(0x0391, 0x03A2),
    range(0x03A3, 0x03AA),
    range(0x03B1, 0x03CA),
    0x03D1,
    range(0x03D5, 0x03D7),
    0x03DA,
    0x03DC,
    0x03DE,
    range(0x03E2, 0x03F2),
    range(0x03F4, 0x03F6),
    range(0x0483, 0x0485),
    0x0487,
    0x0589,
    0x060C,
    range(0x061B, 0x061D),
    0x061F,
    0x0621,
    0x0627,
    0x0640,
    range(0x064B, 0x0656),
    range(0x0660, 0x066D),
    0x0670,
    0x06D4,
    range(0x06F0, 0x06FA),
    0x093D,
    range(0x0951, 0x0953),
    range(0x0964, 0x0970),
    range(0x09E6, 0x09F0),
    range(0x09F4, 0x09F8),
    range(0x0A66, 0x0A70),
    range(0x0AE6, 0x0AF0),
    0x0BAA,
    0x0BB5,
    range(0x0BE6, 0x0BF4),
    range(0x0CE6, 0x0CF0),
    range(0x1040, 0x104A),
    range(0x1735, 0x1737),
    range(0x1801, 0x1804),
    0x1805,
    0x1CD0,
    range(0x1CD2, 0x1CD4),
    range(0x1CD5, 0x1CDB),
    range(0x1CDC, 0x1CDE),
    range(0x1CE0, 0x1CE2),
    range(0x1CE9, 0x1CEB),
    0x1CED,
    range(0x1CF2, 0x1CFA),
    range(0x1EF2, 0x1EFA),
    0x2002,
    range(0x200E, 0x2010),
    0x2011,
    range(0x2013, 0x2015),
    range(0x2018, 0x201B),
    range(0x201C, 0x201F),
    0x2020,
    0x2022,
    0x2026,
    range(0x2032, 0x2034),
    range(0x2039, 0x203B),
    0x2044,
    0x204F,
    0x205A,
    0x20A8,
    range(0x20AA, 0x20AC),
    0x20B4,
    0x20B9,
    0x20DB,
    0x20F0,
    range(0x2190, 0x219A),
    0x2212,
    0x2215,
    range(0x2218, 0x221A),
    0x2299,
    range(0x22C4, 0x22C7),
    range(0x2308, 0x230C),
    range(0x231C, 0x2320),
    0x237C,
    0x23AF,
    0x23D0,
    range(0x2460, 0x2476),
    0x25AF,
    0x25B3,
    0x25B7,
    0x25BD,
    0x25C1,
    0x25CA,
    0x25FB,
    0x262C,
    range(0x2669, 0x2672),
    range(0x2800, 0x2900),
    range(0x2921, 0x2923),
    0x2981,
    0x29BF,
    0x29EB,
    0x2E31,
    0x2E41,
    range(0x3000, 0x3003),
    range(0x3008, 0x3010),
    0xA66F,
    range(0xA78B, 0xA78D),
    range(0xA830, 0xA83A),
    0xA8F1,
    0xA92E,
    0xA9CF,
    range(0xFD3E, 0xFD40),
    0xFDF2,
    0xFDFD,
    0xFE00,
    range(0xFE24, 0xFE27),
    range(0xFE2E, 0xFE30),
    range(0x10100, 0x10103),
    range(0x10107, 0x10134),
    range(0x10137, 0x10140),
    range(0x102E0, 0x102FC),
    0x10AF2,
    range(0x10E60, 0x10E7F),
    range(0x11FD0, 0x11FD2),
    0x11FD3,
    range(0x1D2E0, 0x1D2F4),
    range(0x1EE00, 0x1EE04),
    range(0x1EE05, 0x1EE20),
    range(0x1EE21, 0x1EE23),
    0x1EE24,
    0x1EE27,
    range(0x1EE29, 0x1EE33),
    range(0x1EE34, 0x1EE38),
    0x1EE39,
    0x1EE3B,
    0x1EE42,
    0x1EE47,
    0x1EE49,
    0x1EE4B,
    range(0x1EE4D, 0x1EE50),
    range(0x1EE51, 0x1EE53),
    0x1EE54,
    0x1EE57,
    0x1EE59,
    0x1EE5B,
    0x1EE5D,
    0x1EE5F,
    range(0x1EE61, 0x1EE63),
    0x1EE64,
    range(0x1EE67, 0x1EE6B),
    range(0x1EE6C, 0x1EE73),
    range(0x1EE74, 0x1EE78),
    range(0x1EE79, 0x1EE7D),
    0x1EE7E,
    range(0x1EE80, 0x1EE8A),
    range(0x1EE8B, 0x1EE9C),
    range(0x1EEA1, 0x1EEA4),
    range(0x1EEA5, 0x1EEAA),
    range(0x1EEAB, 0x1EEBC),
    range(0x1EEF0, 0x1EEF2),
]
for e in EXISTING:
    if isinstance(e, int):
        ALLOWED.add(e)
    else:
        ALLOWED |= set(e)


def test_overlaps():
    cp_coverage = defaultdict(list)
    for subset in ListSubsets():
        if subset in SKIP_SUBSETS:
            continue
        for cp in CodepointsInSubset(subset, unique_glyphs=True):
            if cp in ALLOWED:
                continue
            cp_coverage[cp].append(subset)
    failed = []
    for cp, subsets in sorted(cp_coverage.items()):
        if len(subsets) > 1:
            print(f"Codepoint U+{cp:04X} is in multiple subsets: {', '.join(subsets)}")
            failed.append(cp)
    assert not failed



import youseedee
import os
import re
import argparse
import csv

EVERYONE_GETS_THESE = [0x00, 0x0D, 0x20, 0xA0]
OUTPUT_PATH = "Lib/gfsubsets/data/"

blocks = youseedee.database["Blocks.txt"]["reader"]("Blocks.txt")

aliases = {}
with open(
    os.path.join(youseedee.ucd_dir(), "PropertyValueAliases.txt"), "r", newline=""
) as f:
    reader = csv.reader(f, delimiter=";", skipinitialspace=True)
    for row in reader:
        if len(row) != 3:
            continue
        category, short, long_name = row
        if category.strip() != "sc":
            continue
        aliases[long_name.strip()] = short.strip()


def block_to_chars(blockname):
    found = False
    for start, end, name in blocks:
        if name == blockname:
            found = True
            yield from range(start, end + 1)
    if not found:
        raise Exception("Unknown block: %s" % blockname)


def script_to_chars(scriptname):
    scripts = youseedee.database["Scripts.txt"]["reader"]("Scripts.txt")
    script_extensions = youseedee.database["ScriptExtensions.txt"]["reader"](
        "ScriptExtensions.txt"
    )
    if scriptname not in aliases:
        raise ValueError(
            f"Are you sure {scriptname} is a Unicode script? It has no short alias"
        )

    shortname = aliases[scriptname]
    for start, end, script in scripts:
        if script == scriptname:
            yield from range(start, end + 1)

    for start, end, scripts in script_extensions:
        if shortname in scripts:
            yield from list(range(start, end + 1))


def output_chars(todo, filename, args):
    positive = [x for x in todo if x >= 0]
    negative = [-x for x in todo if x < 0]
    codepoints = (set(positive) - set(negative)) | set(EVERYONE_GETS_THESE)
    with open(os.path.join(OUTPUT_PATH, os.path.basename(filename)), "w") as f:
        f.write("# This file was generated by preprocess_namfile.py\n")
        f.write("# It is a list of all codepoints in the Unicode block or script\n")
        f.write(f"# specified in the original file {filename}.\n\n")
        for codepoint in sorted(codepoints):
            data = youseedee.ucd_data(codepoint)
            name = data.get("Name", "")
            if data.get("General_Category") == "Cn" or (not name) and "Age" not in data:
                continue
            f.write("0x%04X %s\n" % (codepoint, name))


def file_to_chars(filename, including=False):
    with open(filename, "r") as f:
        for line in f:
            # A line is either: a comment, a codepoint, a range, a Unicode block
            # or a Unicode script name, an @include statement
            line = line.strip()
            if "@include_only" in line:
                if not including:
                    break
                continue
            if not line or line.startswith("#"):
                continue
            line = re.sub(r"\s+#.*$", "", line)
            sign = 1
            if line[0] == "-":
                line = line[1:]
                sign = -1

            m = re.match(r"^@(block|script|include)\(([^\)]+)\)$", line)

            if m:
                if m.group(1) == "block":
                    yield from [sign * x for x in block_to_chars(m.group(2))]
                elif m.group(1) == "script":
                    yield from [sign * x for x in script_to_chars(m.group(2))]
                else:
                    includefile = os.path.join(os.path.dirname(filename), m.group(2))
                    yield from [
                        sign * x for x in file_to_chars(includefile, including=True)
                    ]
                continue
            # Match a range of codepoints.
            m = re.match(
                r"^(?:U\+|0x)?([0-9A-F]+)\s*\.\.\s*(?:U\+|0x)?([0-9A-F]+)$",
                line,
                re.IGNORECASE,
            )
            if m:
                yield from [
                    sign * x
                    for x in range(int(m.group(1), 16), int(m.group(2), 16) + 1)
                ]
                continue
            # Match a single codepoint.
            m = re.match(r"^(?:U\+|0x)?([0-9A-F]+)(\s|$)", line, re.IGNORECASE)
            if m:
                yield sign * int(m.group(1), 16)
                continue
            else:
                raise Exception("Invalid line: %s" % line)


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Preprocess a set of input glyphset files"
    )
    parser.add_argument(
        "--no-label",
        action="store_true",
        help="Don't emit a representative glyph, just the name",
    )
    parser.add_argument(
        "filenames", metavar="FILENAME", nargs="+", help="The input glyphset files"
    )
    args = parser.parse_args()
    for filename in args.filenames:
        print("Preprocessing " + filename)
        todo = list(file_to_chars(filename))
        if todo:
            output_chars(todo, filename, args)
        else:
            print(" No characters found")


if __name__ == "__main__":
    main()



"""Diff the codepoints supported by two slice files.

Usage:

    python3 scripts/slice_cpdiff file1.nam file2.
"""

from collections import defaultdict
from pathlib import Path
import sys
from typing import Set
import unicodeblock.blocks


def codepoints(slice_file: Path) -> Set[int]:
    codepoints = set()
    with open(slice_file) as f:
        for line in f:
            comment_start = line.find("#")
            if comment_start != -1:
                line = line[:comment_start]
            line = line.strip()
            if not line.startswith("codepoints: "):
                continue
            codepoints.add(int(line[line.index(" ") :]))
    return codepoints


def path_to_file(maybe: str) -> Path:
    path = Path(maybe)
    assert path.is_file(), f"{path} must be a file"
    return path


def list(tag: str, source: str, codepoints: Set[int]):
    for cp in sorted(codepoints):
        print(f"{tag} {source} 0x{cp:04x}")


def list_blocks(prefix: str, codepoints: Set[int]):
    blocks = defaultdict(int)
    for cp in codepoints:
        blocks[unicodeblock.blocks.of(chr(cp))] += 1
    blocks = sorted([(count, block) for (block, count) in blocks.items()], reverse=True)
    for count, block in blocks:
        print(f"{prefix}{block} {count}")


def main(argv):
    assert len(argv) == 3, "Must have exactly two arguments"

    nam1 = path_to_file(argv[1])
    nam2 = path_to_file(argv[2])
    nam1cp = codepoints(nam1)
    nam2cp = codepoints(nam2)

    identical = nam1cp & nam2cp
    only1 = nam1cp - nam2cp
    only2 = nam2cp - nam1cp

    list("only", nam1.name, only1)
    list("only", nam2.name, only2)

    print()
    print(f"{len(identical)} match")
    print(f"{len(only1)} only in {nam1.name}")
    list_blocks("  ", only1)
    print(f"{len(only2)} only in {nam2.name}")
    list_blocks("  ", only2)


if __name__ == "__main__":
    main(sys.argv)
